#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of Youba (Youpi back end)
# Copyright (C) 2011 AlphaK - http://www.alphak.net/
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

PROGRAM_NAME = 'Youba'
VERSION = '1.1'
BUILD = '20110808'

import logging, SocketServer, sys, os, settings
import threading
from classes import CommandPart, Action, WaitMode, Motor, Rotation, Unit
from classes import SpecialAction, InstructionChainModifier, InstructionChain
from backend import Backend
log = logging.getLogger('listener')
log.addHandler(logging.StreamHandler())


#==============================================================================
# 
#==============================================================================

class Youba(SocketServer.BaseRequestHandler):
	"""
	This class deals with commands sent by clients/frontends

	It is instantiated once per connection to the server, and must
	override the handle() method to implement communication to the
	client.
	"""
	
	SEP_SA = '!!'
		

	def handle(self):
		
		# self.request is the TCP socket connected to the client
		self.data = ''
		while self.data.endswith('\n') == False:
			self.data += self.request.recv(1024)
		self.data = self.data.strip()
		log.info('Command sent from [%s]: [%s]' % (self.client_address[0], self.data))
		command = self.data
		
              
		try:
					
			# Determine if the command is a specialAction or an instructionChainModifier
			if command.endswith(self.SEP_SA):
				# This is a specialAction: (I|C|R)!!
				specialAction = SpecialAction.checkSanity(command.replace(self.SEP_SA, ''))
				if specialAction == SpecialAction.INFORMATION:
					c = threadBackend.readCounters()
					self.request.send('OK %s\n' % c)
				elif specialAction == SpecialAction.COUNTERRESET:
					threadBackend.resetCounters()
					self.request.send('OK\n')
				elif specialAction == SpecialAction.REINIT:
					threadBackend.initRobot()
					self.request.send('OK\n')
				else:
					# We should never get here
					pass
				
				# Log relevant info
				log.info('OK - command from %s received correctly' % self.client_address[0])
				
			else:
				# This is an instructionChainModifier: [action!waitMode!]instructionSet
				instructionChainModifier = InstructionChainModifier(command)
				# We need exclusive access to the instructionChainModifier
				lock.acquire()
				instructionChain.update(instructionChainModifier)
				lock.release()
			
				# Send status to client and log
				self.request.send('OK\n')
				log.info('OK - command from %s received correctly' % self.client_address[0])
				
            
		except Exception,e:
			# This exception is raised mainly in case of incorrect command syntax, but can catch any other unexpected instruction
			self.request.send('KO\n')
			log.warning('KO - %s' % e)

		
		
#==============================================================================
# Main part of the program, checks settings, starts and stops the server
#==============================================================================

if __name__ == '__main__':

	# Settings validation
	for key in ['LOG_FILE', 'LOG_LEVEL', 'DUMP_FILE', 'SERVER_IP', 'SERVER_PORT', 'PARALLEL_PORT']:
		if not settings.__dict__.has_key(key):
			print 'File settings.py is incomplete! Key [%s] is missing!' % key
			sys.exit(1)

	HOST, PORT = settings.SERVER_IP, settings.SERVER_PORT
	
	# Logging facilities
	SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
	LEVELS = {
		'debug': logging.DEBUG,
		'info': logging.INFO,
		'warning': logging.WARNING,
		'error': logging.ERROR,
		'critical': logging.CRITICAL
	}
	level = LEVELS.get(settings.LOG_LEVEL, logging.NOTSET)
	logging.basicConfig(level=level, filename='%s/%s' % (SCRIPT_DIR, settings.LOG_FILE), filemode='w')
	
	# Initializes instruction chain and server
	instructionChain = InstructionChain()
	server = None
	
	# Creates an initializes backend thread and a lock for synchronization
	lock = threading.Lock()
	try:
		threadBackend = Backend('Backend', instructionChain, lock)
	except Exception,e:
		log.error(e)
		sys.exit(1)
		
	threadBackend.start()

	# Creates the server, binding host and port with read settings
	try:
		server = SocketServer.TCPServer((HOST, PORT), Youba)
	except Exception,e:
		log.error('Unable to start server: %s' % e)
		sys.exit(2)

	# Server was started successfully
	log.info('Server %s v%s build %s started.' % (PROGRAM_NAME, VERSION, BUILD))
	
	# Activates the server, this will keep running until interrupted with Ctrl-C, any other signal, or uncaught exception
	try:
		server.serve_forever()
	except:
		log.info('Process interrupted.')
	finally:
		log.info('Stopping server, please wait...')
		threadBackend.stop()
		log.info('Server %s stopped.' % PROGRAM_NAME)

